#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <sys/param.h>
#include "internal.h"
#include "fsys.h"

/* 
 * The fsys interface abstracts the file system from the rest of oct. 
 * Mapping from <cell,view,facet,version> to file names is done here 
 * and here alone.  Management of file permissions and locking is done 
 * here as well.  See the fsys interface description for the 
 * individual functions
 * 
 * For this simple implementation on top of UNIX, the cell and view 
 * are nested directories, with the facets being files.  The name of 
 * the file is "cell/view/facet;version", where cell may be a unix 
 * path name its own right.  No file locking is done and 
 * the commit primitives are ignored.  The cell and view directories 
 * are created at flush time, so creation of bogus directories because 
 * of typo's in the cell and view name are avoided.  Because we dont 
 * create the directories at the beginning, we cant simply uniquely 
 * indentify files by there inode number.  So we indentify them by the 
 * inode number of the  directory containing the cell and then the name 
 * of the cell view facet and version.  This scheme can be fooled by a cell
 * that is actually a symbolic link to a cell at a different location, 
 * so care should be taken.
 * 
 */
#define alloc(type) ((type *) malloc(sizeof(type)))

extern char *sprintf(), *strcpy(), *index(), *rindex();
extern char *strcat(), *strncpy();
static char *tilde_expand();

extern int errno;
extern int sys_nerr;
extern char *sys_errlist[];

static char fsys_error_buf[1024];
char *fsys_error_message = fsys_error_buf;

#define sys_msg() (errno < sys_nerr ? sys_errlist[errno] : "Unknown error")

struct desc {
    dev_t device;
    ino_t inode;
    char *file_name;
    char *tail;
};


/* search backward from ptr in buffer for a / */
    
#define FIND_NEXT_SLASH(buffer, ptr)\
    while( *--ptr != '/' && ptr != buffer) {\
    }\
    if (*ptr != '/') {\
	sprintf(fsys_error_buf, "Panic: mangled file_desc in files.c");\
	return 0;\
    }


/*ARGSUSED*/
fsys_commit(desc)
char *desc;
{
    return 1;
}

/*ARGSUSED*/
fsys_lock(desc)
char *desc;
{
    return 1;
}

/*ARGSUSED*/
fsys_unlock(desc)
char *desc;
{
    return 1;
}

fsys_print_key(file, c_desc)
FILE *file;
char *c_desc;
{
    struct desc *desc = (struct desc *) c_desc;
    
    fprintf(file, "dev = %ld, inode = %d, file = %s, tail = %s",
	    desc->device, desc->inode, desc->file_name, desc->tail);
}

/* 
 * return a FILE pointer to the file associated with the descripter c_desc.
 * Create the cell and view directories, if necessary
 */

FILE *fsys_open(c_desc, mode)
char *c_desc;
char *mode;
{
    struct desc *desc = (struct desc *) c_desc;
    char *file_name = desc->file_name;
    FILE *stream;
    char *ptr, *cell_dir, *view_dir;

    ptr = file_name + strlen(file_name);

    FIND_NEXT_SLASH(file_name, ptr);
    
    view_dir = ptr;

    FIND_NEXT_SLASH(file_name, ptr);
    
    cell_dir = ptr;

    *cell_dir = '\0';
    
    if (access(file_name, F_OK) != 0 && mkdir(file_name, 0777) != 0) {
	sprintf(fsys_error_buf, "Cannot open %s: %s", file_name, sys_msg());
	*cell_dir = '/';
	return NULL;
    }
    
    *cell_dir = '/';
    *view_dir = '\0';
    
    if (access(file_name, F_OK) != 0 && mkdir(file_name, 0777) != 0) {
	sprintf(fsys_error_buf, "Cannot open %s: %s", file_name, sys_msg());
	*view_dir = '/';
	return NULL;
    }
    
    *view_dir = '/';
    
    stream = fopen(file_name, mode);
    
    if (stream == NULL) {
	sprintf(fsys_error_buf, "Cannot open %s: %s", file_name, sys_msg());
    }
    
    return stream;
}

extern int errno;
extern int sys_nerr;
extern char *sys_errlist[];

#define sys_msg() (errno < sys_nerr ? sys_errlist[errno] : "Unknown error")

static char *search_path[] = {".", "..", 0};

#define ABSOLUTE_PATH(name) ((name)[0] == '/' || (name)[0] == '~')

/* 
 * Figure out the filename associated with user_facet.  If the nameing 
 * is relative (c_parent_desc != nil), then the name is looked up 
 * relative to the directory containing the parent cell or in the 
 * parent cell directory itself.
 */

int
fsys_make_key(user_facet, c_file_desc, c_parent_desc)
struct octFacet *user_facet;
char **c_file_desc;
char *c_parent_desc;
{
    char *last_slash;
    static char desc_buffer[MAXPATHLEN];
    static struct desc desc;
    struct stat stat_buf;

    if (ABSOLUTE_PATH(user_facet->cell)) {
	if (!canonicalize(desc_buffer, user_facet->cell, "/")) {
	    sprintf(fsys_error_buf, "Can't expand the file name %s",
		    user_facet->cell);
	    return 0;
	}

	last_slash = rindex(desc_buffer, '/');
	
	if (last_slash == NIL(char)) {
	    sprintf(fsys_error_buf,
		    "Panic: %s didn't expand into a absolute pathname.\n",
		    user_facet->cell);
	    return 0;
	}
    } else if (c_parent_desc != (char *) 0) {
	struct desc *parent_desc = (struct desc *) c_parent_desc;
	char *parent_tail = index(parent_desc->tail+1, '/');
	char *cwd = parent_desc->file_name;
	char **path;

	*parent_tail = '\0';  /* turn the facet filename (parent_desc->file_name and also cwd)
				 into the cell directory name */
	
	for (path = search_path; *path != (char *) 0; path++) {
	    if (canonicalize(desc_buffer, *path, cwd)) {
		sprintf(desc_buffer + strlen(desc_buffer), "/%s", user_facet->cell);
		if (access(desc_buffer, F_OK) == 0) {
		    break;
		}
	    }
	}

	if (*path == (char *) 0) {
	    /*
	     * no path found with cell already existing,
	     * so we search for a writable path
	     */
	    for (path = search_path; *path != (char *) 0; path++) {
		if (canonicalize(desc_buffer, *path, cwd)) {
		    if (access(desc_buffer, F_OK) == 0) {
			break;
		    }
		}
	    }
	    if (*path == (char *) 0) {
		*parent_tail = '/';
		sprintf(fsys_error_buf, "Can't find cell `%s'",
			oct_facet_name(user_facet));
		return 0;
	    }
	    sprintf(desc_buffer + strlen(desc_buffer), "/%s", user_facet->cell);
	}
	
	*parent_tail = '/'; /* repair what we did to parent_desc->file_name above */
    } else {
	sprintf(desc_buffer, "./%s", user_facet->cell);
    }

    last_slash = rindex(desc_buffer, '/');
    *last_slash = '\0';
    if (stat(desc_buffer, &stat_buf) != 0) {
	sprintf(fsys_error_buf, "Can't stat directory %s: %s", desc_buffer,
		sys_msg());
	return 0;
    }
    *last_slash = '/';

    desc.device = stat_buf.st_dev;
    desc.inode = stat_buf.st_ino;
    sprintf(desc_buffer + strlen(desc_buffer), "/%s/%s;%s", user_facet->view,
	    user_facet->facet,
	   (user_facet->version == NIL(char) ? "" : user_facet->version));
    desc.file_name = desc_buffer;

    desc.tail = last_slash;
    
    *c_file_desc = (char *) &desc;
    return 1;
}

/* return the permissions that the user has on the facet file */

octStatus
fsys_resolve(c_file_desc)
char *c_file_desc;
{
    char *tail, *head;
    int mode;
    char desc_buffer[MAXPATHLEN];
    struct desc *file_desc = (struct desc *) c_file_desc;
	
    
    strcpy(desc_buffer, file_desc->file_name);
    
    if (access(desc_buffer, F_OK) == 0) {
	mode = FSYS_EXISTS;
	if (access(desc_buffer, R_OK) == 0) {
	    mode |= FSYS_READABLE;
	}
	if (access(desc_buffer, W_OK) == 0) {
	    mode |= FSYS_WRITABLE;
	}
	return mode;
    }

    /*
     * The facet file doesn't exist, make sure it is at least creatable,
     */
    
    tail = desc_buffer + strlen(desc_buffer);
    
    FIND_NEXT_SLASH(desc_buffer, tail);
    *tail = '\0';
    
    if (access(desc_buffer, F_OK) == 0) {
	if (access(desc_buffer, W_OK|X_OK) == 0) {
	    return FSYS_CREATABLE;
	} else {
	    *tail = '/';
	    sprintf(fsys_error_buf, "Can't create facet file `%s': %s",
		    desc_buffer, sys_msg());
	    return 0;
	}
    }

    /*
     * the view directory doesn't exist, make sure it can be created
     */
    
    FIND_NEXT_SLASH(desc_buffer, tail);
    *tail = '\0';
    
    if (access(desc_buffer, F_OK) == 0) {
	if (access(desc_buffer, W_OK|X_OK) == 0) {
	    return FSYS_CREATABLE;
	} else {
	    *tail = '/';
	    sprintf(fsys_error_buf, "Can't create view directory file `%s': %s",
		    desc_buffer, sys_msg());
	    return 0;
	}
    }

    /*
     * the cell directory doesn't exist, make sure it can be created
     */
    
    tail = rindex(desc_buffer, '/');
    
    if (tail == (char *) 0) {
	head = ".";
    } else if (tail == desc_buffer) {
	head = "/";
    } else {
	head = desc_buffer;
	*tail = '\0';
    }
    
    if (access(head, F_OK) == 0) {
	if (access(head, W_OK|X_OK) == 0) {
	    return FSYS_CREATABLE;
	} else {
	    *tail = '/';
	    sprintf(fsys_error_buf, "Can't create cell directory `%s': %s",
		    desc_buffer, sys_msg());
	    return 0;
	}
    }
    
    /*
     * Can't even find the directory the cell is in, give up
     */
    
    sprintf(fsys_error_buf, "Can't find the cell directory `%s': %s", head,
	    sys_msg());
    return 0;
}

char *getwd();
/* 
 * fsys_make_key makes a static copy of the key information, save_key 
 * makes a new copy on the heap so it wont get overwritten on the next 
 * call to make_key.  This is an optimization: fsys_make_key has to be called
 * every time we check to see if a cell is in memory, but 
 * fsys_save_key only needs to be called when reading the cell for the 
 * first time.  We save all the string copying and the expensive call 
 * to getwd.
 */
fsys_save_key(key, copy)
char *key;
char **copy;
{
    struct desc *desc, *orig = (struct desc *) key;
    char *full_path, *tail;

    if (orig->file_name[0] != '/') {
	char buffer[MAXPATHLEN];
	int buffer_length;
	if (getwd(buffer) == NIL(char)) {
	    sprintf(fsys_error_buf,"Can't determine current directory");
	    return 0;
	}
	buffer_length = strlen(buffer);
	full_path =
	  malloc((unsigned) (buffer_length+strlen(orig->file_name)+2));
	if (full_path == NIL(char)) {
	    sprintf(fsys_error_buf,"Out of memory");
	    return 0;
	}
	sprintf(full_path, "%s/%s", buffer, orig->file_name);
	tail = full_path + buffer_length + 1 + (orig->tail - orig->file_name);
    } else {
	full_path = strsave(orig->file_name);
	tail = full_path + (orig->tail - orig->file_name);
    }
    
    desc = alloc(struct desc);
    if (desc == NIL(struct desc)) {
	FREE(full_path);
	sprintf(fsys_error_buf, "Out of Memory");
	return 0;
    }
    desc->device = orig->device;
    desc->inode = orig->inode;
    desc->file_name = full_path;
    desc->tail = tail;
    *copy = (char *) desc;
    return 1;
}

fsys_free_key(c_key)
char *c_key;
{
    struct desc *key = (struct desc *) c_key;
    FREE(key->file_name);
    FREE(key);
    return 1;
}

fsys_compare_key(c_a, c_b)
char *c_a, *c_b;
{
    struct desc *a = (struct desc *) c_a;
    struct desc *b = (struct desc *) c_b;

    if (a->device != b->device) {
	return a->device - b->device;
    } else if (a->inode != b->inode) {
	return a->inode - b->inode;
    } else {
	return strcmp(a->tail, b->tail);
    }
}
	
fsys_hash_key(c_a, modulus)
char *c_a;
int modulus;
{
    struct desc *a = (struct desc *) c_a;
    return st_strhash(a->tail, modulus);
}
/* 
 * canonicalize just does tilde expand and maps the empty path ("") to 
 * the current directory
 */
static
canonicalize(buffer, path, cwd)
char *buffer;
char *path;
char *cwd;
{
    char *ptr = buffer;

    switch (path[0]) {
    case '~':
	if (tilde_expand(ptr, path) == NIL(char)) {
	    return 0;
	} else {
	    return 1;
	}
    case '\0':
	strcpy(buffer, "./");
	return 1;
    case '/':
	strcpy(buffer, path);
	return 1;
    default:
	sprintf(buffer, "%s/%s", cwd, path);
	return 1;
    }
}

/* 
 * Expand names of the form ~user into the users home directory. A 
 * cache of previous mappings is kept in order to speed things up
 */

#include "st.h"

static st_table *name_table = NIL(st_table);

static char *
tilde_expand(buffer, input)
char *buffer;
char *input;
{
    char *end, *dir, *name;
    int name_length;
    struct passwd *entry;

    if (input[0] != '~') {
	strcpy(buffer, input);
	return buffer + strlen(buffer);
    }
    
    end = index(input,'/');
    if (end == (char *) 0) {
	end = input + strlen(input);
    }

    name_length = end - input - 1;

    strncpy(buffer, input+1, name_length);
    buffer[name_length] = '\0';

    if (!name_table) {
	name_table =
	  st_init_table_with_params(strcmp, st_strhash,ST_DEFAULT_INIT_TABLE_SIZE,
				    ST_DEFAULT_MAX_DENSITY, ST_DEFAULT_GROW_FACTOR, 1);
    }

    if (!name_table || !st_lookup(name_table, buffer, &dir)) {
	
	if (name_length == 0) {
	    entry = getpwuid(getuid());
	} else {
	    entry = getpwnam(buffer);
	}

	if (entry == (struct passwd *) 0) {
	    return NIL(char);
	}

	dir = strsave(entry->pw_dir);
	if (!dir) {
	    return NIL(char);
	}

	name = strsave(buffer);
	if (!name) {
	    FREE(dir);
	    return NIL(char);
	}

	st_insert(name_table, name, dir);
    }
		  
    strcpy(buffer, dir);
    strcat(buffer, end);
    return buffer + strlen(buffer);
}


#ifdef TEST

#define HELP\
"p - change path, x - change cwd. c - canonicalize name. ?,h - this message\n"
main()
{
    char path[MAXPATHLEN], cell[MAXPATHLEN], cwd[MAXPATHLEN],
         output[MAXPATHLEN];
    char c[2];
    int retval;
    struct octFacet facet;
    
    facet.view = "view";
    facet.facet = "facet";
    facet.version = (char *) 0;
    facet.cell = cell;
    strcpy(path, "");
    strcpy(cwd, ".");

    printf(HELP);
    for(;;) {
	printf("Command: ");
	if (scanf("%1s", c) != 1) exit(0);
	switch (c[0]) {
	case 'x' :
	    printf("New cwd: ");
	    if (scanf("%s", cwd) != 1) {
		exit(-1);
	    }
	    break;
	case 'p' :
	    printf("New path (in quotes): ");
	    if (scanf("%*[ \t\n]\"%[^\"]\"", path) != 1) {
		printf("Bad format for path\n");
		scanf("%*[^\n]");
	    }
	    break;
	case 'c' :
	    printf("Canonicalize: ");
	    if (scanf("%s", cell) != 1) {
		exit(-1);
	    }
	    retval = canonicalize(output, &facet, path, cwd);
	    printf("Canonicalize(%s, %s, %s) returns %s (%d)\n", facet.cell,
		   path, cwd, output, retval);
	    break;
	case '?' :
	case 'h' :
	    printf(HELP);
	    break;
	}
    }
}
#endif
 
